import numpy as np
from gym import utils
from gym.envs.mujoco import mujoco_env
import os
import random

DEFAULT_CAMERA_CONFIG = {
    'distance': 4.0,
}


class HalfCheetahEnv(mujoco_env.MujocoEnv, utils.EzPickle):
    def __init__(self,
                 xml_file='half_cheetah.xml',
                 forward_reward_weight=1.0,
                 ctrl_cost_weight=0.1,
                 reset_noise_scale=0.1,
                 exclude_current_positions_from_observation=True):
        utils.EzPickle.__init__(**locals())

        self._forward_reward_weight = forward_reward_weight

        self._ctrl_cost_weight = ctrl_cost_weight

        self._reset_noise_scale = reset_noise_scale

        self._exclude_current_positions_from_observation = (
            exclude_current_positions_from_observation)

        dir_path = os.path.dirname(os.path.realpath(__file__))
        mujoco_env.MujocoEnv.__init__(self, ('%s/assets/'+xml_file) % dir_path, 5)

    def control_cost(self, action):
        control_cost = self._ctrl_cost_weight * np.sum(np.square(action))
        return control_cost

    def step(self, action):
        x_position_before = self.sim.data.qpos[0]
        self.do_simulation(action, self.frame_skip)
        x_position_after = self.sim.data.qpos[0]
        x_velocity = ((x_position_after - x_position_before)
                      / self.dt)

        ctrl_cost = self.control_cost(np.clip(action, -1, 1))

        forward_reward = self._forward_reward_weight * x_velocity

        observation = self._get_obs()
        reward = forward_reward - ctrl_cost
        done = False
        info = {
            'x_position': x_position_after,
            'x_velocity': x_velocity,

            'reward_run': forward_reward,
            'reward_ctrl': -ctrl_cost
        }

        return observation, reward, done, info

    def _get_obs(self):
        position = self.sim.data.qpos.flat.copy()
        velocity = self.sim.data.qvel.flat.copy()

        if self._exclude_current_positions_from_observation:
            position = position[1:]

        observation = np.concatenate((position, velocity)).ravel()
        return observation

    def set_observation(self, observation):
        self.set_state(observation[0:len(self.init_qpos)], observation[len(self.init_qpos):len(self.init_qpos)+len(self.init_qvel)])

    def reset_model(self):
        noise_low = -self._reset_noise_scale
        noise_high = self._reset_noise_scale

        qpos = self.init_qpos + self.np_random.uniform(
            low=noise_low, high=noise_high, size=self.model.nq)
        qvel = self.init_qvel + self._reset_noise_scale * self.np_random.randn(
            self.model.nv)

        self.set_state(qpos, qvel)

        observation = self._get_obs()
        return observation

    def viewer_setup(self):
        for key, value in DEFAULT_CAMERA_CONFIG.items():
            if isinstance(value, np.ndarray):
                getattr(self.viewer.cam, key)[:] = value
            else:
                setattr(self.viewer.cam, key, value)

class HalfCheetahFeasibilityEnv(HalfCheetahEnv):
    def __init__(self,
                 xml_file='half_cheetah.xml',
                 forward_reward_weight=1.0,
                 ctrl_cost_weight=0.1,
                 reset_noise_scale=0.1,
                 exclude_current_positions_from_observation=True, demos=[]):

        self.demos = demos
        self.step_ = 0
        self.current_traj = None
        super(HalfCheetahFeasibilityEnv, self).__init__(xml_file, forward_reward_weight, ctrl_cost_weight,
                 reset_noise_scale, exclude_current_positions_from_observation)

    def step(self, action):
        self.do_simulation(action, self.frame_skip)

        observation = self._get_obs()
        if self.current_traj is None:
            reward = 0
        else:
            reward = -np.linalg.norm(self.current_traj[self.step_+1] - observation, ord=2)

        self.step_ += 1
        if self.current_traj is not None and self.step_ >= len(self.current_traj)-1:
            done = True
        else:
            done = False
        info = {
        }

        return observation, reward, done, info

    def reset_model(self):
        self.step_ = 0
        self.current_traj = self.demos[random.sample(range(len(self.demos)), 1)[0]]
        qpos = self.current_traj[0][0:len(self.init_qpos)]
        qvel = self.current_traj[0][len(self.init_qpos):len(self.init_qvel)+len(self.init_qpos)]
        self.set_state(qpos, qvel)

        observation = self._get_obs()
        return observation

    def set_initial_state(self, demo):
        self.step_ = 0
        self.current_traj = demo
        qpos = self.current_traj[0][0:len(self.init_qpos)]
        qvel = self.current_traj[0][len(self.init_qpos):len(self.init_qvel)+len(self.init_qpos)]
        self.set_state(qpos, qvel)

        observation = self._get_obs()
        return observation

